#include <iterator>
#include "parser.h"

bool is_color(string s) {
    return(s[0] >= 'a' && s[0] <= 'z');
}

bool is_size(string s) {
    return(s[0] >= '0' && s[0] <= '9');
}

void Context::add(string con) {
    if (con == "root") {
        return;
    }

    if (con == "B") {
        if (!b) {
            b = true;
            cons.push_back(con);
        }
    }
    else if ((con == "EM") && !s) {
        em = !em;

        if (em) {
            cons.push_back(con);
        }
        else {
            remove_vector(con);
        }
    }
    else if (con == "I") {
        if (!i) {
            i = true;
            cons.push_back(con);
        }
    }
    else if (con == "PL") {
        s  = false; remove_vector(CONS_S);
        em = false; remove_vector(CONS_EM);
        i  = false; remove_vector(CONS_I);
        b  = false; remove_vector(CONS_B);
        tt = false; remove_vector(CONS_TT);
        u  = 0;     remove_vector(CONS_U);
    }
    else if (con == "S") {
        if (!s) {
            s  = true;  cons.push_back(CONS_S);
            em = false; remove_vector(CONS_EM);
        }
    }
    else if (con == "TT") {
        if (!tt) {
            tt = true;
            cons.push_back(CONS_TT);
        }
    }
    else if (con == "U") {
        if (u < 3) {
            ++u; cons.push_back(CONS_U);
        }
    }
    else if (is_color(con)) {
        cons.erase(remove_if(cons.begin(), cons.end(), is_color), cons.end());
        cons.push_back(con);
        color = con[0];
    }
    else if (is_size(con)) {
        cons.erase(remove_if(cons.begin(), cons.end(), is_size), cons.end());
        cons.push_back(con);
        size = con[0];
    }
}

void Token::print_context(){
    cout << '[';
    copy(cons.begin(), cons.end(), ostream_iterator<string>(cout, " "));
    cout << ']';
}

static void set_context(token tok, vector<token> stk) {
    for (vector<token>::iterator i = stk.begin(); i != stk.end(); ++i) {
        tok->context.add((*i)->str);
    }
    sort(tok->cons.begin(), tok->cons.end());
}

void parse(FILE* in, token_list& tokens){
    string str;
    char   ch;
    bool   in_tag = false;
    bool   is_end = false;
    bool   is_ws  = false;
    int    num_tt = 0;
    token  tok;

    vector<token> stk;

    while ((ch = fgetc(in)) != EOF) {
        if (isspace(ch)) {
            if (!is_ws || (num_tt > 0)) {
                is_ws = true;
                str += ch;
            }
        }
        else if (ch == '<') {
            in_tag = true;
            is_end = false;

            if (!str.empty()) {
                tok = token(new Token(str));

                set_context(tok, stk);

                tokens.push_back(tok);
                str.erase();
            }
        }
        else if (ch == '>') {
            in_tag = false;
            tok    = token(new Token(str, true));

            if (is_end) {
                stk.pop_back();

                if (str == "TT") {
                    --num_tt;
                }
            }
            else {
                if (str == "TT") {
                    ++num_tt;
                }

                stk.push_back(tok);
            }

            str.erase();
        }
        else if (in_tag && (ch == '/')) {
            is_end = true;
        }
        else {
            is_ws = false;
            str  += ch;
        }
    }

    if (!str.empty()) {
        tokens.push_back(token(new Token(str)));
    }
}

