#ifndef _PARSER_H_
#define _PARSER_H_

#include <algorithm>
#include <cstdio>
#include <list>
#include <string>
#include <stack>
#include <vector>
#include "ptr_wrapper.h"

#define NUM_CONTEXTS 20

#define I_B     0
#define I_EM    1
#define I_S     2
#define I_I     3
#define I_TT    4
#define I_U     5
#define I_SIZE  6
#define I_COLOR 7

const string CONS_B  = string("B");
const string CONS_EM = string("EM");
const string CONS_S  = string("S");
const string CONS_I  = string("I");
const string CONS_TT = string("TT");
const string CONS_U  = string("U");
const string CONS_PL = string("PL");

struct Token;

typedef PtrWrapper<Token>             token;
typedef list<token>                   token_list;
typedef list<token>::iterator         token_iterator;
typedef list<token>::reverse_iterator token_r_iterator;

typedef vector<string>                   string_list;
typedef vector<string>::iterator         string_iterator;
typedef vector<string>::reverse_iterator string_r_iterator;

typedef stack<string> string_stack;

struct Context {
    string_list cons;

    bool b;
    bool em;
    bool i;
    bool s;
    bool tt;
    bool pl;

    int u;
    char size;
    char color;

    bool pl_able() {
        return !b && !em && !i && !s && !tt && u == 0;
    }

    Context() {
        b = em = i = s = tt = false;
        u = 0;
        size  = 0;
        color = 0;
    }

    bool contains(string con) {
        return(find(cons.begin(), cons.end(), con) != cons.end());
    }

    void add(string con);

    void remove_vector(string con) {
        cons.erase(::remove(cons.begin(), cons.end(), con), cons.end());
    }
};

struct Token {
    string str;
    bool   is_tag;
    bool   is_open;
    uint   depth;

    Token* parent;
    Token* next_sibling;
    Token* prev_sibling;
    Token* pl_tok;

    Context      context;
    string_list& cons;
    token_list   children;


    Token(string str) : cons(context.cons) {
        init(str, false);
    }

    Token(string str, bool is_tag) : cons(context.cons) {
        init(str, is_tag);
    }

    Token(string str, bool is_tag, bool is_open) : cons(context.cons) {
        init(str, is_tag);

        this->is_open = is_open;
    }

    ~Token() {
        cout << "destroy: " << str << endl;
    }

    void init(string str, bool is_tag) {
        this->str    = str;
        this->is_tag = is_tag;
        parent       = NULL;
        prev_sibling = NULL;
        next_sibling = NULL;
        depth        = 0;
        pl_tok       = NULL;
    }

    void print_context();

    bool contains_context(string con) {
        return context.contains(con);
    }

    char getColor() {
        return str[0];
    }

    char getSize() {
        return str[0];
    }

    bool hasColor() {
        return context.color > 0;
    }

    bool hasSize() {
        return context.size > 0;
    }

    bool isColor() {
        return context.color > 0;
    }

    bool isSize() {
        return context.size > 0;
    }

    token_iterator cbegin() {
        return children.begin();
    }

    token_iterator cend() {
        return children.end();
    }

    bool good_for_pl(Token* tok) {
        int num = 0;
        for (uint i = 0; i < cons.size(); ++i) {
            if (tok->contains_context(cons[i])) {
                ++num;
            }
        }

        return(num > 1);
    }
};

void add_context(Context& context, string st);
void parse(FILE* in, token_list& root);

#endif // _PARSER_H_
